require("mod-gui")

local function delete_chunks() 
	if global.chunks and #global.chunks > 0 then
		local this_chunk = global.chunks[#global.chunks]
		local surface = game.surfaces[this_chunk.idx]
		local area = {{this_chunk.x*32, this_chunk.y*32}, {this_chunk.x*32+32, this_chunk.y*32+32}}
		local entities = surface.find_entities(area)
		for _, entity in pairs(entities) do
			script.raise_event(defines.events.on_entity_died, {entity=entity}) --raise event so that mods can do their stuff with the entities
		end
		for _, entity in pairs(entities) do
			if entity.valid then entity.destroy() end
		end
		--game.players[1].print(this_chunk.x .. this_chunk.y)
		surface.delete_chunk({this_chunk.x, this_chunk.y})
		global.chunks[#global.chunks] = nil
		if #global.chunks == 1 then --the function tends to miss some chunks, this prevents that
			local chunks = {}
			for chunk in surface.get_chunks() do
				chunk.idx = surface.index
				chunks[#chunks+1] = chunk
			end
			global.chunks = chunks
		end
		if global.chunks == nil then --make sure that if the previous code made it nil the event is unregistered, since the on load assumes it has to be unregistered if it's nil
			script.on_event(defines.events.on_tick, nil)
		end
	else
		global.chunks = nil
		script.on_event(defines.events.on_tick, nil)
	end
end

-- GUI --
local function kill_gui(player)
	local button_flow = mod_gui.get_button_flow(player)
	local frame_flow = mod_gui.get_frame_flow(player)		
	if button_flow["new-game-plus-toggle-config"] then
		button_flow["new-game-plus-toggle-config"].destroy()
	end
	if frame_flow["new-game-plus-config-frame"] then
		frame_flow["new-game-plus-config-frame"].destroy()
	end
	if frame_flow["new-game-plus-config-more-frame"] then
		frame_flow["new-game-plus-config-more-frame"].destroy()
	end
end

local function generate_resource_options(resource_name, config_resource_table, freq_options, size_options, richn_options)
	if resource_name == "lithia-water" then return end --belongs to ground-water autoplace control (bob's mod)
	config_resource_table.add{
		type = "label",
		caption = {"autoplace-control-names." .. resource_name},
		name = "new-game-plus-config-" .. resource_name .. "-label",
	}
	local resource_freq = config_resource_table.add{
		type = "drop-down",
		name = "new-game-plus-config-" .. resource_name .. "-freq",
	}
	resource_freq.items = freq_options
	resource_freq.selected_index = 3
	local resource_size = config_resource_table.add{
		type = "drop-down",
		name = "new-game-plus-config-" .. resource_name .. "-size",
	}
	resource_size.items = size_options
	resource_size.selected_index = 4
	local resource_richn = config_resource_table.add{
		type = "drop-down",
		name = "new-game-plus-config-" .. resource_name .. "-richn",
	}
	resource_richn.items = richn_options
	resource_richn.selected_index = 3
end

local function make_advanced_settings_gui(config_more_frame)
	local config_more_flow = config_more_frame.add{
		type = "flow",
		name = "new-game-plus-config-more-flow",
		direction = "vertical"
	}
	config_more_flow.style.max_on_row = 1
	local config_more_option_table = config_more_flow.add{
		type = "table",
		name = "new-game-plus-config-more-option-table",
		colspan = 2
	}
	config_more_option_table.add{
		type = "label",
		caption = {"gui-map-generator.pollution"},
		name = "new-game-plus-pollution-label",
	}
	config_more_option_table.add{
		type = "checkbox",
		name = "new-game-plus-pollution-checkbox",
		state = true,
	}
	config_more_option_table.add{
		type = "label",
		caption = {"gui-map-generator.evolution"},
		name = "new-game-plus-pollution-enabled-label",
	}
	config_more_option_table.add{
		type = "checkbox",
		name = "new-game-plus-evolution-checkbox",
		state = true,
	}
	config_more_option_table.add{
		type = "label",
		caption = {"gui-map-generator.enemy-expansion-group-tile"},
		name = "new-game-plus-enemy-expansion-label",
	}
	config_more_option_table.add{
		type = "checkbox",
		name = "new-game-plus-enemy-expansion-checkbox",
		state = true,
	}
	config_more_flow.add{
		type = "label",
		caption = {"gui-map-generator.recipes-and-technology-group-tile"},
		style = "caption_label_style",
		name = "new-game-plus-difficulty-title-label",
	}
	-- difficulty options
	local config_more_option_difficulty_table = config_more_flow.add{
		type = "table",
		name = "new-game-plus-config-more-option-difficulty-table",
		colspan = 2
	}
	local difficulty_options = {{"recipe-difficulty.normal"}, {"recipe-difficulty.expensive"}}
	config_more_option_difficulty_table.add{
		type = "label",
		caption = {"gui-map-generator.recipe-difficulty"},
		name = "new-game-plus-recipe-difficulty-label",
	}
	local recipe_difficulty = config_more_option_difficulty_table.add{
		type = "drop-down",
		name = "new-game-plus-recipe-difficulty-drop-down",
	}
	recipe_difficulty.items = difficulty_options
	recipe_difficulty.selected_index = 1
	config_more_option_difficulty_table.add{
		type = "label",
		caption = {"gui-map-generator.technology-difficulty"},
		name = "new-game-plus-technology-difficulty-label",
	}
	local technology_difficulty = config_more_option_difficulty_table.add{
		type = "drop-down",
		name = "new-game-plus-technology-difficulty-drop-down",
	}
	technology_difficulty.items = difficulty_options
	technology_difficulty.selected_index = 1
	config_more_option_difficulty_table.add{
		type = "label",
		caption = {"gui-map-generator.technology-price-multiplier"},
		name = "new-game-plus-technology-multiplier-label",
	}
	local technology_multiplier = config_more_option_difficulty_table.add{
		type = "textfield",
		name = "new-game-plus-technology-multiplier-textfield",
	}
	technology_multiplier.text = "1"
	technology_multiplier.style.maximal_width = 50
end

local function make_resource_settings_gui(config_frame)
	local config_resource_scroll_pane = config_frame.add{
		type = "scroll-pane",
		name = "new-game-plus-config-resource-scroll-pane",
	}
	config_resource_scroll_pane.style.maximal_height = 400
	config_resource_scroll_pane.style.visible = true
	local config_resource_table = config_resource_scroll_pane.add{
		type = "table",
		name = "new-game-plus-config-resource-table",
		colspan = 4
	}
	config_resource_table.add{
		type = "label",
		caption = " ",
		name = "new-game-plus-config-resource-empty-label",
	}
	config_resource_table.add{
		type = "label",
		caption = {"gui-map-generator.frequency"},
		style = "caption_label_style",
		name = "new-game-plus-config-resource-freq-label",
	}
	config_resource_table.add{
		type = "label",
		caption = {"gui-map-generator.size"},
		style = "caption_label_style",
		name = "new-game-plus-config-resource-size-label",
	}
	config_resource_table.add{
		type = "label",
		caption = {"gui-map-generator.richness"},
		style = "caption_label_style",
		name = "new-game-plus-config-resource-richn-label",
	}
	local freq_options = {{"frequency.very-low"}, {"frequency.low"}, {"frequency.normal"}, {"frequency.high"}, {"frequency.very-high"}}
	local size_options = {{"size.none"}, {"size.very-low"}, {"size.low"}, {"size.normal"}, {"size.high"}, {"size.very-high"}}
	local richn_options = {{"richness.very-low"}, {"richness.low"}, {"richness.normal"}, {"richness.high"}, {"richness.very-high"}}
	--water stuff
	config_resource_table.add{
		type = "label",
		caption = {"gui-map-generator.water"},
		name = "new-game-plus-config-water-label",
	}
	local resource_freq = config_resource_table.add{ --terrain segmentation
		type = "drop-down",
		name = "new-game-plus-config-water-freq",
	}
	resource_freq.items = freq_options
	resource_freq.selected_index = 3
	local resource_size = config_resource_table.add{  --"water" in the map gen settings
		type = "drop-down",
		name = "new-game-plus-config-water-size",
	}
	resource_size.items = size_options --first option should actually be size.only-starting-area but that localized string is waaaay too long
	resource_size.selected_index = 4
	config_resource_table.add{
		type = "label",
		caption = " ",
		name = "new-game-plus-config-resource-empty-label-2",
	}
	--biters
	if game.entity_prototypes["biter-spawner"] and game.entity_prototypes["biter-spawner"].autoplace_specification and game.entity_prototypes["spitter-spawner"] and game.entity_prototypes["spitter-spawner"].autoplace_specification then
		generate_resource_options("enemy-base", config_resource_table, freq_options, size_options, richn_options)
	end
	--resources
	for _, resource in pairs(game.entity_prototypes) do
		if resource.type == "resource" and resource.autoplace_specification then
			generate_resource_options(resource.name, config_resource_table, freq_options, size_options, richn_options)
		end
	end
	--starting area
	config_resource_table.add{
		type = "label",
		caption = {"gui-map-generator.starting-area"},
		name = "new-game-plus-config-starting-area-label",
	}
	config_resource_table.add{
		type = "label",
		caption = " ",
		name = "new-game-plus-config-resource-empty-label-3",
	}
	local resource_size = config_resource_table.add{
		type = "drop-down",
		name = "new-game-plus-config-starting-area-size",
	}
	resource_size.items = {{"size.very-low"}, {"size.low"}, {"size.normal"}, {"size.high"}, {"size.very-high"}}
	resource_size.selected_index = 3
	config_resource_table.add{
		type = "label",
		caption = " ",
		name = "new-game-plus-config-resource-empty-label-4",
	}
end

local function make_basic_settings_gui(config_frame)
	local config_option_table = config_frame.add{
		type = "table",
		name = "new-game-plus-config-option-table",
		colspan = 2
	}
	config_option_table.add{
		type = "label",
		name = "new-game-plus-reset-research-label",
		caption = {"gui.new-game-plus-reset-research-caption"},
	}
	config_option_table.add{
		type = "checkbox",
		name = "new-game-plus-reset-research-checkbox",
		state = false,
	}
	config_option_table.add{
		type = "label",
		name = "new-game-plus-peaceful-mode-label",
		caption = {"gui-map-generator.peaceful-mode"},
	}
	config_option_table.add{
		type = "checkbox",
		name = "new-game-plus-peaceful-mode-checkbox",
		state = false,
	}
	config_option_table.add{
		type = "label",
		name = "new-game-plus-seed-label",
		caption = {"gui.new-game-plus-seed-caption"},
	}
	local seed_textfield = config_option_table.add{
		type = "textfield",
		name = "new-game-plus-seed-textfield",
	}
	seed_textfield.text = "0"
	config_option_table.add{
		type = "label",
		name = "new-game-plus-width-label",
		caption = {"gui.new-game-plus-width-caption"},
	}
	local width_textfield = config_option_table.add{
		type = "textfield",
		name = "new-game-plus-width-textfield",
	}
	width_textfield.text = "0"
	config_option_table.add{
		type = "label",
		name = "new-game-plus-height-label",
		caption = {"gui.new-game-plus-height-caption"},
	}
	local height_textfield = config_option_table.add{
		type = "textfield",
		name = "new-game-plus-height-textfield",
	}
	height_textfield.text = "0"
end

local function regen_gui(player)
	kill_gui(player)
	--toggle button
	local button_flow = mod_gui.get_button_flow(player)
	local button = button_flow.add{
		type = "button",
		name = "new-game-plus-toggle-config",
		style = mod_gui.button_style,
		tooltip = {"gui.new-game-plus-toggle-tooltip", {"gui.new-game-plus-config-caption"}},
		caption = {"gui.new-game-plus-toggle-config-caption"}
	}
	button.style.visible = true
	-- general gui frame setup --
	local frame_flow = mod_gui.get_frame_flow(player)
	local config_frame = frame_flow.add{
		type = "frame",
		name = "new-game-plus-config-frame",
		direction = "vertical"
	}
	config_frame.style.visible = false
	local config_frame_title_table = config_frame.add{
		type = "table",
		name = "new-game-plus-config-frame-title-table",
		colspan = 2
	}
	local config_frame_title = config_frame_title_table.add{
		type = "label",
		name = "new-game-plus-config-frame-title-label",
		caption = {"gui.new-game-plus-config-caption"},
	}
	config_frame_title.style.font = "default-frame"
	config_frame_title.style.right_padding = 5
	config_frame_title_table.add{
		type = "button",
		name = "new-game-plus-more-options",
		style = mod_gui.button_style,
		tooltip = {"gui.new-game-plus-toggle-tooltip", {"gui-map-generator.advanced-tab-title"}},
		caption = {"gui-map-generator.advanced-tab-title"}
	}
	local config_more_frame = frame_flow.add{
		type = "frame",
		caption = {"gui-map-generator.advanced-tab-title"},
		name = "new-game-plus-config-more-frame",
		direction = "vertical"
	}
	config_more_frame.style.visible = false
	
	--make gui sections
	make_basic_settings_gui(config_frame)
	make_resource_settings_gui(config_frame)
	make_advanced_settings_gui(config_more_frame)
	
	-- start button at the bottom
	local start_button = config_frame.add{
		type = "button",
		name = "new-game-plus-start-button",
		--style = mod_gui.button_style,
		tooltip = {"gui.new-game-plus-start-button-tooltip"},
		caption = {"gui.new-game-plus-start-button-caption"}
	}
	start_button.style.font_color = { r=1, g=0.75, b=0.22}
end

script.on_event(defines.events.on_rocket_launched, function(event)
	if global.rocket_launched then return end --don't regen gui if this is not the first rocket launch
	if event.rocket.get_item_count("satellite") > 0 then
		for _, player in pairs(game.players) do
			regen_gui(player)
		end
		global.rocket_launched = true --the rocket has now been launched, so the gui has to be regened on player created and config changed
	end
end)

-- WOLRD GEN --
local function textfield_to_uint(textfield)
	if textfield.text and tonumber(textfield.text) and tonumber(textfield.text) >= 0 and tonumber(textfield.text) <= 4294967295 then
		return tonumber(textfield.text)
	else
		return false
	end
end

local function change_map_settings(player)
	local frame_flow = mod_gui.get_frame_flow(player)
	local map_settings = game.map_settings
	local config_more_options = frame_flow["new-game-plus-config-more-frame"]["new-game-plus-config-more-flow"]["new-game-plus-config-more-option-table"]
	map_settings.pollution.enabled = config_more_options["new-game-plus-pollution-checkbox"].state
	map_settings.enemy_evolution.enabled = config_more_options["new-game-plus-evolution-checkbox"].state
	map_settings.enemy_expansion.enabled = config_more_options["new-game-plus-enemy-expansion-checkbox"].state
	local config_more_options_difficulty = frame_flow["new-game-plus-config-more-frame"]["new-game-plus-config-more-flow"]["new-game-plus-config-more-option-difficulty-table"]
	local recipe_difficulty_options = {defines.difficulty_settings.recipe_difficulty.normal, defines.difficulty_settings.recipe_difficulty.expensive}
	game.difficulty_settings.recipe_difficulty = recipe_difficulty_options[config_more_options_difficulty["new-game-plus-recipe-difficulty-drop-down"].selected_index]
	local technology_difficulty_options = {defines.difficulty_settings.technology_difficulty.normal, defines.difficulty_settings.technology_difficulty.expensive}
	game.difficulty_settings.technology_difficulty = config_more_options_difficulty["new-game-plus-technology-difficulty-drop-down"].selected_index
	local technology_price_multiplier = textfield_to_uint(config_more_options_difficulty["new-game-plus-technology-multiplier-textfield"])
	if technology_price_multiplier and technology_price_multiplier > 0 then
		game.difficulty_settings.technology_price_multiplier = technology_price_multiplier
	else
		player.print({"msg.new-game-plus-start-invalid-technology-multiplier"})
		return false
	end
	return true
end

local function make_map_gen_settings(player)
	local frame_flow = mod_gui.get_frame_flow(player)
	local config_options = frame_flow["new-game-plus-config-frame"]["new-game-plus-config-option-table"]
	local map_gen_settings = {}
	local seed = textfield_to_uint(config_options["new-game-plus-seed-textfield"])
	if seed and seed == 0 then
		map_gen_settings.seed = math.random(0, 4294967295)
	elseif seed then
		map_gen_settings.seed = seed
	else 
		player.print({"msg.new-game-plus-start-invalid-seed"})
		return nil
	end
	local width = textfield_to_uint(config_options["new-game-plus-width-textfield"])
	if width then
		map_gen_settings.width = width
	else
		player.print({"msg.new-game-plus-start-invalid-width"})
		return nil
	end
	local height = textfield_to_uint(config_options["new-game-plus-height-textfield"])
	if height then
		map_gen_settings.height = height
	else
		player.print({"msg.new-game-plus-start-invalid-height"})
		return nil
	end
	if config_options["new-game-plus-peaceful-mode-checkbox"].state then
		map_gen_settings.peaceful_mode = true
	else
		map_gen_settings.peaceful_mode = false
	end		
	local freq_options = {"very-low", "low", "normal", "high", "very-high"}
	local size_options = {"none", "very-small", "small", "medium", "big", "very-big"}
	local richn_options = {"very-poor", "poor", "regular", "good", "very-good"}
	local resource_table = frame_flow["new-game-plus-config-frame"]["new-game-plus-config-resource-scroll-pane"]["new-game-plus-config-resource-table"]
	--water stuff
	map_gen_settings.terrain_segmentation = freq_options[resource_table["new-game-plus-config-water-freq"].selected_index]
	map_gen_settings.water = size_options[resource_table["new-game-plus-config-water-size"].selected_index]
	--starting area
	local starting_area_options = {"very-small", "small", "medium", "big", "very-big"}
	map_gen_settings.starting_area = starting_area_options[resource_table["new-game-plus-config-starting-area-size"].selected_index]
	-- resources
	local autoplace_controls_mine = {}
	for _, resource in pairs(game.entity_prototypes) do
		if resource.type == "resource" and resource.autoplace_specification and resource.name ~= "lithia-water" then
			autoplace_controls_mine[resource.name] = {
				frequency = freq_options[resource_table["new-game-plus-config-" .. resource.name .. "-freq"].selected_index],
				size = size_options[resource_table["new-game-plus-config-" .. resource.name .. "-size"].selected_index],
				richness = richn_options[resource_table["new-game-plus-config-" .. resource.name .. "-richn"].selected_index]
			}
		end
	end
	if game.entity_prototypes["biter-spawner"] and game.entity_prototypes["biter-spawner"].autoplace_specification and game.entity_prototypes["spitter-spawner"] and game.entity_prototypes["spitter-spawner"].autoplace_specification then
		autoplace_controls_mine["enemy-base"] = {
		frequency = freq_options[resource_table["new-game-plus-config-enemy-base-freq"].selected_index],
		size = size_options[resource_table["new-game-plus-config-enemy-base-size"].selected_index],
		richness = richn_options[resource_table["new-game-plus-config-enemy-base-richn"].selected_index]
		}
	end
	map_gen_settings.autoplace_controls = autoplace_controls_mine
	return map_gen_settings
end

local function generate_new_world(player)
	-- MAP GEN SETTINGS
	local map_gen_settings = make_map_gen_settings(player)
	if not map_gen_settings then return end
	-- MAP SETTINGS
	if not change_map_settings(player) then return end
	-- teleport players to new surface --
	local nauvis_plus = game.create_surface("Nauvis plus " .. global.next_nauvis_number, map_gen_settings)
	for i, player in pairs(game.players) do
		player.teleport({1, i}, nauvis_plus)
		player.force.chart(nauvis_plus, {{player.position.x - 200, player.position.y - 200}, {player.position.x + 200, player.position.y + 200}})
	end 
	--set spawn
	for _, force in pairs(game.forces) do
		force.set_spawn_position({0,0}, nauvis_plus)
	end
	--set whether the tech needs to be reset
	local frame_flow = mod_gui.get_frame_flow(player)
	if frame_flow["new-game-plus-config-frame"]["new-game-plus-config-option-table"]["new-game-plus-reset-research-checkbox"].state then
		global.tech_reset = true
	else
		global.tech_reset = false
	end
	--kill the gui, it's not needed and should not exist in the background
	for _, player in pairs(game.players) do
		kill_gui(player)
	end
	--destroy the other surfaces
	for _,surface in pairs(game.surfaces) do
		if surface.name == "nauvis" then --can't delete nauvis
			local chunks = {}
			for chunk in surface.get_chunks() do
				chunk.idx = surface.index
				chunks[#chunks+1] = chunk
			end
			global.chunks = chunks
			--delete 1 chunk per tick
			script.on_event(defines.events.on_tick,	delete_chunks)
		elseif not surface.name:find("Factory floor") and surface.name ~= "Nauvis plus " .. global.next_nauvis_number then --don't delete factorissimo stuff or the new surface
			game.delete_surface(surface)
		end
	end
	global.next_nauvis_number = global.next_nauvis_number + 1
	global.rocket_launched = false	--we act like no rocket has been launched, so that we can do the whole "launch a satellite" thing again
end

script.on_event({defines.events.on_gui_click}, function(event)
	local player = game.players[event.player_index]
	local frame_flow = mod_gui.get_frame_flow(player)
	if event.element.name == "new-game-plus-toggle-config" then
		frame_flow["new-game-plus-config-frame"].style.visible = not frame_flow["new-game-plus-config-frame"].style.visible
		frame_flow["new-game-plus-config-more-frame"].style.visible = false
	elseif event.element.name == "new-game-plus-more-options" then
		frame_flow["new-game-plus-config-more-frame"].style.visible = not frame_flow["new-game-plus-config-more-frame"].style.visible
	elseif event.element.name == "new-game-plus-start-button" then
		if not player.admin then
			player.print({"msg.new-game-plus-start-admin-restriction"})
			return
		end
		generate_new_world(player)
	end
end)

local function make_grass_bridge(from, to, surface) --makes a two tile wide north-south bridge if there is water
	local tile_table = {}
	for y=from, to do
		local tile = surface.get_tile(0, y)
		if tile.name:find("water") then
			table.insert(tile_table,{name="grass", position={0, y}})
		end
		local tile = surface.get_tile(1, y)
		if tile.name:find("water") then
			table.insert(tile_table,{name="grass", position={1, y}})
		end
	end 
	surface.set_tiles(tile_table)
end

script.on_event(defines.events.on_chunk_generated, function(event) --prevent island spawns (hopefully)
	if global.next_nauvis_number == 1 then return end
	local this_nauvis_number = global.next_nauvis_number - 1
	local surface = event.surface
	if surface.name == "Nauvis plus " .. this_nauvis_number then 
		local chunk_area = event.area
		if chunk_area.left_top.x == 0 and chunk_area.left_top.y == 0 and chunk_area.right_bottom.x == 32 and chunk_area.right_bottom.y == 32 then
			make_grass_bridge(0, 32, surface)
			if global.tech_reset then --tech reset on chunk gen because then the spilling works correctly
				for _, force in pairs(game.forces) do
					for _, tech in pairs(force.technologies) do
						tech.researched = false;
						force.set_saved_technology_progress(tech, 0)
					end
				end
			end
		end
		if chunk_area.left_top.x == 0 and chunk_area.left_top.y == 32 and chunk_area.right_bottom.x == 32 and chunk_area.right_bottom.y == 64 then
			make_grass_bridge(33, 64, surface)
		end
		if chunk_area.left_top.x == 0 and chunk_area.left_top.y == 64 and chunk_area.right_bottom.x == 32 and chunk_area.right_bottom.y == 96 then
			make_grass_bridge(65, 95, surface)
		end
	end
end)

script.on_load(function()
	if global.chunks then
		script.on_event(defines.events.on_tick, delete_chunks)
	end
end)

script.on_configuration_changed(function() --regen gui in case a mod added/removed resources, only if a rocket has been launched
	if global.rocket_launched then
		for _, player in pairs(game.players) do
			regen_gui(player)
		end
	end
end)

script.on_event(defines.events.on_player_created, function(event) --create gui for joining player, only if a rocket has been launched
	if global.rocket_launched then
		regen_gui(game.players[event.player_index])
	end
end)

script.on_init(function() 
	global.next_nauvis_number = 1
end)
local function regen_gui_com(event)
	regen_gui(game.players[event.player_index])
end
commands.add_command("regen_gui", "Regen gui", regen_gui_com)